from .gps_data import GpsPoints  # noqa
from .gps_data import PoiPoints  # noqa
from .gps_data import load_gps_points  # noqa
from .gps_data import load_poi_points  # noqa
from .io import save  # noqa
from . import plot_utils  # noqa
from . import raster_analysis  # noqa
from . import time_analysis  # noqa
from . import utils  # noqa



import os
import pytest

from .fixtures.gps_points import *  # noqa
from .fixtures.poi_points import *  # noqa


@pytest.fixture
def gpkg_filepath(tmpdir):
    tmpdir_str = tmpdir.strpath
    filename = "test.gpkg"
    return os.path.join(tmpdir_str, filename)



import pytest

import numpy as np

import gps_data_analyzer as gda


_res_X = [
    [-0.15, -0.15, -0.15, -0.15, -0.15],
    [-0.025, -0.025, -0.025, -0.025, -0.025],
    [0.1, 0.1, 0.1, 0.1, 0.1],
    [0.225, 0.225, 0.225, 0.225, 0.225],
    [0.35, 0.35, 0.35, 0.35, 0.35],
]
_res_Y = [
    [0.85, 0.975, 1.1, 1.225, 1.35],
    [0.85, 0.975, 1.1, 1.225, 1.35],
    [0.85, 0.975, 1.1, 1.225, 1.35],
    [0.85, 0.975, 1.1, 1.225, 1.35],
    [0.85, 0.975, 1.1, 1.225, 1.35],
]
_res_values = [
    [0.0, 0.190739, 0.353963, 0.333906, 0.14622],
    [0.190739, 0.544946, 0.787424, 0.685123, 0.333906],
    [0.353963, 0.787424, 1.0, 0.787424, 0.353963],
    [0.333906, 0.685123, 0.787424, 0.544946, 0.190739],
    [0.14622, 0.333906, 0.353963, 0.190739, 0.0],
]


def test_extent():
    extent = gda.raster_analysis.Extent(0, 1, 2, 3)

    assert extent.xmin == extent[0] == 0
    assert extent.xmax == extent[1] == 1
    assert extent.ymin == extent[2] == 2
    assert extent.ymax == extent[3] == 3
    assert [i for i in extent] == [0, 1, 2, 3]


def test_extent_border():
    extent = gda.raster_analysis.Extent(0, 1, 2, 3, 0.1)

    assert extent.inner_xmin == 0
    assert extent.inner_xmax == 1
    assert extent.inner_ymin == 2
    assert extent.inner_ymax == 3
    assert [i for i in extent] == [-0.1, 1.1, 1.9, 3.1]

    extent.reset_border(0.2)

    assert extent.inner_xmin == 0
    assert extent.inner_xmax == 1
    assert extent.inner_ymin == 2
    assert extent.inner_ymax == 3
    assert [i for i in extent] == [-0.2, 1.2, 1.8, 3.2]


def test_simple_heatmap(simple_gps_data):
    res = gda.raster_analysis.heatmap(simple_gps_data, mesh_size=0.1)

    assert np.equal(res.X, [[0, 0], [0.2, 0.2]]).all()
    assert np.equal(res.Y, [[1, 1.2], [1, 1.2]]).all()
    assert np.allclose(res.values, [[0, 1], [1, 0]])


def test_heatmap_no_normalization(simple_gps_data):
    res = gda.raster_analysis.heatmap(simple_gps_data, mesh_size=0.1, normalize=False)

    assert np.equal(res.X, [[0, 0], [0.2, 0.2]]).all()
    assert np.equal(res.Y, [[1, 1.2], [1, 1.2]]).all()
    assert np.allclose(res.values, [[3.042058, 3.103057], [3.103057, 3.042058]])


def test_heatmap_border(simple_gps_data):
    res = gda.raster_analysis.heatmap(simple_gps_data, mesh_size=0.1, border=0.15)

    assert np.allclose(res.X, _res_X,)
    assert np.allclose(res.Y, _res_Y)
    assert np.allclose(res.values, _res_values)


def test_heatmap_nx_ny(simple_gps_data):
    res = gda.raster_analysis.heatmap(simple_gps_data, nx=5, ny=5, border=0.15)

    assert np.allclose(res.X, _res_X,)
    assert np.allclose(res.Y, _res_Y)
    assert np.allclose(res.values, _res_values)


def test_heatmap_args(simple_gps_data):
    # Test conflicting arguments
    for k in ["mesh_size", "x_size", "y_size", "nx", "ny"]:
        with pytest.raises(ValueError):
            gda.raster_analysis.heatmap(simple_gps_data, **{k: 1})

    with pytest.raises(ValueError):
        gda.raster_analysis.heatmap(simple_gps_data, mesh_size=1, nx=1, ny=1)

    with pytest.raises(ValueError):
        gda.raster_analysis.heatmap(simple_gps_data, mesh_size=1, x_size=1)

    with pytest.raises(ValueError):
        gda.raster_analysis.heatmap(simple_gps_data, x_size=1, nx=1, ny=1)

    with pytest.raises(ValueError):
        gda.raster_analysis.heatmap(simple_gps_data, y_size=1, nx=1, ny=1)

    with pytest.raises(ValueError):
        gda.raster_analysis.heatmap(simple_gps_data, x_size=1, y_size=1, nx=1, ny=1)

    with pytest.raises(ValueError):
        gda.raster_analysis.heatmap(
            simple_gps_data, mesh_size=1, x_size=1, y_size=1, nx=1, ny=1
        )

    # Test negative values
    with pytest.raises(ValueError):
        gda.raster_analysis.heatmap(simple_gps_data, mesh_size=-1)

    with pytest.raises(ValueError):
        gda.raster_analysis.heatmap(simple_gps_data, x_size=-1, y_size=1)

    with pytest.raises(ValueError):
        gda.raster_analysis.heatmap(simple_gps_data, x_size=1, y_size=-1)

    with pytest.raises(ValueError):
        gda.raster_analysis.heatmap(simple_gps_data, nx=-1, ny=1)

    with pytest.raises(ValueError):
        gda.raster_analysis.heatmap(simple_gps_data, nx=1, ny=-1)

    with pytest.raises(ValueError):
        gda.raster_analysis.heatmap(simple_gps_data, kernel_size=-1)

    with pytest.raises(ValueError):
        gda.raster_analysis.heatmap(simple_gps_data, kernel_cut=-1)



import logging

import numpy as np
import pandas as pd
from scipy import spatial


def compute_rest_time(gps_data, radius):
    """Compute the duration during which the track stays in a given radius of each
    point.

    Args:
        gps_data (:py:class:`~gps_data_analyzer.gps_data.PoiPoints`): The data used for
            computation.
        radius (float): The radius in which the rest time is computed around each point.

    Returns:
        ``pandas.Series``: The rest time around each point.
    """

    # TODO: need optimization and cleaning.

    def _t_inter(current, i1, i2, max_radius, geom="geometry", t="datetime"):
        d_i1 = i1[geom].distance(current[geom])
        d_i2 = i2[geom].distance(current[geom])
        t_i1 = i1[t]
        t_i2 = i2[t]
        dt = max(t_i1, t_i2) - min(t_i1, t_i2)
        dd = abs(d_i1 - d_i2)
        if dd == 0:
            return pd.Timedelta(0)
        else:
            return min(1.0, abs(d_i1 - max_radius) / dd) * dt

    def _process_one_pt(num, points, max_radius, logger=logging):
        logger.debug("{}: {}".format(num, points))
        data = gps_data.data
        pts = np.array(points)
        pts.sort()
        pos_i = np.argwhere(pts == num)[0][0]
        diff_not_one = (pts[1:] - pts[:-1]) != 1

        current = data.loc[num]

        # TODO: make a function for inf and sup parts since the only difference is the
        # order of diff_not_one and the limits for label_inf_m1 and label_sup_p1

        # Inf part
        if len(diff_not_one[:pos_i]) > 0:
            diff_not_one[0] = True
            pos_skip_inf = pos_i - np.argmax(np.flip(diff_not_one[:pos_i]))
            label_inf = pts[pos_skip_inf]
            label_inf_m1 = max(0, pts[pos_skip_inf] - 1)
            inf = data.loc[label_inf]
            inf_m1 = data.loc[label_inf_m1]
            dt_inf = current["datetime"] - inf["datetime"]
            t_inf_inter = dt_inf + _t_inter(current, inf, inf_m1, max_radius)
            logger.debug("data:\n{}".format(data.loc[[num, label_inf, label_inf_m1]]))
            logger.debug(
                "distances = {}".format(
                    data.loc[[label_inf, label_inf_m1], "geometry"].distance(
                        current["geometry"]
                    )
                )
            )
        else:
            t_inf_inter = pd.Timedelta(0)

        # Sup part
        if len(diff_not_one[pos_i:]) > 0:
            diff_not_one[-1] = True
            pos_skip_sup = pos_i + np.argmax(diff_not_one[pos_i:])
            label_sup = pts[pos_skip_sup]
            label_sup_p1 = min(data.index.max(), pts[pos_skip_sup] + 1)
            sup = data.loc[label_sup]
            sup_p1 = data.loc[label_sup_p1]
            dt_sup = sup["datetime"] - current["datetime"]
            t_sup_inter = dt_sup + _t_inter(current, sup, sup_p1, max_radius)
            logger.debug("data:\n {}".format(data.loc[[num, label_sup, label_sup_p1]]))
            logger.debug(
                "distances = {}".format(
                    data.loc[[label_sup, label_sup_p1], "geometry"].distance(
                        current["geometry"]
                    )
                )
            )
        else:
            t_sup_inter = pd.Timedelta(0)

        logger.debug("t_inf_inter = {}".format(t_inf_inter))
        logger.debug("t_sup_inter = {}".format(t_sup_inter))

        return t_inf_inter, t_sup_inter

    # Get the closest points of each points
    points = np.c_[gps_data.x.ravel(), gps_data.y.ravel()]
    tree = spatial.KDTree(points)
    points = tree.data
    in_radius_pts = tree.query_ball_point(points, radius)

    # Get the times when the track leave the circle with radius = radius
    t_min = []
    t_max = []
    for num, i in enumerate(in_radius_pts):
        t1, t2 = _process_one_pt(num, i, radius)
        t_min.append(t1)
        t_max.append(t2)

    times = pd.DataFrame({"dt_min": t_min, "dt_max": t_max}, index=gps_data.index)

    # Compute total time
    duration = times["dt_min"] + times["dt_max"]

    # Convert time in seconds
    return duration.apply(pd.Timedelta.total_seconds)



import numpy as np


def haversine(lat1, lon1, lat2, lon2, to_radians=True, earth_radius=6371e3):
    """Vectorized haversine function

    Slightly modified version: of http://stackoverflow.com/a/29546836/2901002

    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees or in radians)

    All (lat, lon) coordinates must have numeric dtypes and be of equal length.

    Args:
        lat1 (``numpy.array``): The latitude coordinates of the first point.
        lon1 (``numpy.array``): The longitude coordinates of the first point.
        lat2 (``numpy.array``): The latitude coordinates of the second point.
        lon2 (``numpy.array``): The longitude coordinates of the second point.
        to_radians (bool, optional): Convert coordinates in radians.
        earth_radius (float, optional): The earth radius used for computations.

    Returns:
        ``numpy.array``: The distances betwen each couple of points.
    """
    if to_radians:
        lat1 = np.radians(lat1)
        lon1 = np.radians(lon1)
        lat2 = np.radians(lat2)
        lon2 = np.radians(lon2)

    a = np.power(np.sin((lat2 - lat1) / 2.0), 2) + (
        np.cos(lat1) * np.cos(lat2) * np.power(np.sin((lon2 - lon1) / 2.0), 2)
    )

    return earth_radius * 2.0 * np.arcsin(np.sqrt(a))



import pytest

import pandas as pd

import gps_data_analyzer as gda


@pytest.fixture
def simple_gps_raw_data():
    x = [0, 0.1, 0.2]
    y = [1.0, 1.1, 1.2]
    z = [0, 100, 200]
    t = ["2019/06/16-09:55:34", "2019/06/16-09:55:57", "2019/06/16-09:58:12"]
    return x, y, z, t


@pytest.fixture
def simple_gps_df(simple_gps_raw_data):
    x, y, z, t = simple_gps_raw_data
    df = pd.DataFrame({"x": x, "y": y, "z": z, "t": t})
    return df


@pytest.fixture
def simple_gps_data(simple_gps_df):
    return gda.GpsPoints(simple_gps_df, x_col="x", y_col="y", z_col="z", time_col="t")



import cartopy.crs as ccrs
import cartopy.io.img_tiles as cimgt
from matplotlib.colors import ListedColormap
import matplotlib.pyplot as plt
import numpy as np


def create_transparent_cmap(name="rainbow"):
    """Create a colormap with transparency for the lowest value.

    Returns:
        ``matplotlib.colors.ListedColormap``: The colormap.
    """
    default_cmap = plt.get_cmap(name)
    cmap = default_cmap(np.arange(default_cmap.N))
    cmap[:, -1] = np.linspace(0, 1, default_cmap.N)
    return ListedColormap(cmap)


def add_annotated_points(ax, points, **kwargs):
    """Add annotated points to the given axis.

    Args:
        ax (``matplotlib.pyplot.Axes``): The axis object to update.
        points (:py:class:`~gps_data_analyzer.gps_data.PoiPoints`): The points to add to
            the axis.
        kwargs: The given kwargs will be passed to :func:`matplotlib.pyplot.annotate`.

    Note:
        The `points` object must be a ``pandas.DataFrame`` with the following columns:

            * x
            * y
            * name (optional): labels are taken if not provided
            * horizontalalignment (or ha) (optional): the default value is 'center'
            * verticalalignment (or va) (optional): the default value is 'center'
            * fontsize (optional): the default value is 12
            * x_offset (optional): the default value is 0
            * y_offset (optional): the default value is 0
    """

    pts = points.copy()

    # If name column is missing, set IDs as names
    if "name" not in pts.columns:
        pts["name"] = pts.index.astype(str)

    # Set default arguments for arrowprops and bbox
    arrowprops = kwargs.pop("arrowprops", dict(facecolor="black", shrink=0.05))
    bbox = kwargs.pop("bbox", dict(facecolor="sandybrown", alpha=0.5, boxstyle="round"))

    # Create a matplotlib transform object for the coordinate system
    proj = ax.projection._as_mpl_transform(ax)

    # Add annotations
    for num, row in pts.iterrows():
        ax.annotate(
            row["name"],
            xy=(row["geometry"].x, row["geometry"].y),
            xytext=(
                row["geometry"].x + row.get("x_offset", 0),
                row["geometry"].y + row.get("y_offset", 0)
            ),
            arrowprops=arrowprops,
            xycoords=proj,
            fontsize=row.get("fontsize", 12),
            ha=row.get("ha", row.get("horizontalalignment", "center")),
            va=row.get("va", row.get("verticalalignment", "center")),
            bbox=bbox,
            **kwargs
        )


def setup_axis(
    ax=None,
    extent=None,
    projection=None,
    background=False,
    zoom=10
):
    """Setup a ``matplotlib.pyplot.Axes`` instance.

    Args:
        ax (``matplotlib.pyplot.Axes``, optional): The axis object to update.
        extent (:py:class:`~gps_data_analyzer.raster_analysis.Extent` or :py:obj:`list`\
            of :py:obj:`float`): The extent to set.
        projection (:py:class:`cartopy.crs.Projection`, optional): The projection of the
            axis (:py:class:`~cartopy.crs.PlateCarree` by default).
        background (bool or :py:class:`cartopy.io.img_tiles.GoogleWTS`, optional): If
            true, a default background is added using Google Satellite. If a
            :obj:`~cartopy.io.img_tiles.GoogleWTS` object is given, it is used.
        zoom (int, mandatory if :py:obj:`background` is not :py:obj:`None`): The zoom
            value used to generate the background.
    """

    # Define projection
    if projection is None:
        projection = ccrs.PlateCarree()

    # Create the figure instance
    if ax is None:
        fig = plt.figure()

        # Create a GeoAxes in the tile's projection.
        ax = fig.add_subplot(1, 1, 1, projection=projection)
    else:
        fig = None

    # Limit the extent of the map to the min/max coords
    if extent is not None:
        ax.set_extent(extent, crs=projection)

    # Create a terrain background instance
    if background is True:
        background = cimgt.GoogleTiles(style="satellite")  # pragma: no cover

    # Add the background data
    if background is not False:
        ax.add_image(background, zoom)

    return fig, ax


def plot(
    data,
    var=None,
    vmin=None,
    vmax=None,
    ax=None,
    show=True,
    projection=None,
    background=False,
    zoom=None,
    annotations=None,
    annotation_kwargs=None,
    border=0,
    extent=None,
    **kwargs
):
    """Plot data with background and annotations.

    Args:
        data (:py:class:`~gps_data_analyzer.gps_data._GpsBase`): The object to plot.
        var (str, optional): The column used as plotted value.
        vmin (float, optional): The lower clipping value.
        vmax (float, optional): The upper clipping value.
        ax (``matplotlib.pyplot.Axes``, optional): The axis object to update.
        show (bool, optional): If true, call :func:`plt.show` else return the figure
            and axis objects.
        projection (:py:class:`cartopy.crs.Projection`, optional): The projection of the
            axis (:py:class:`~cartopy.crs.PlateCarree` by default).
        background (bool or :py:class:`cartopy.io.img_tiles.GoogleWTS`, optional): If
            true, a default background is added using Google Satellite. If a
            :obj:`~cartopy.io.img_tiles.GoogleWTS` object is given, it is used.
        zoom (int, mandatory if :py:obj:`background` is not :py:obj:`None`): The zoom
            value used to generate the background.
        annotations (:py:class:`~gps_data_analyzer.gps_data.PoiPoints`, optional): The
            points used to annotate the figure.
        annotation_kwargs (dict): The kwargs passed to
            :func:`~gps_data_analyzer.plot_utils.add_annotated_points`.
        border (float, optional): The extra border around the data (not used if
            :py:attr:`extent` is given).
        extent (:py:class:`~gps_data_analyzer.raster_analysis.Extent` or :py:obj:`list`\
            of :py:obj:`float`): The extent to set.
        kwargs: The given kwargs will be passed to :func:`pandas.DataFrame.plot`.

    Returns:
        The figure and axis if :py:attr:`show` is False, :py:obj:`None` otherwise.
    """

    # Limit the extent of the map to the min/max coords
    if extent is None:
        xmin, ymin, xmax, ymax = data.sindex.bounds
        extent = (xmin - border, xmax + border, ymin - border, ymax + border)

    # Setup axis
    fig, ax = setup_axis(
        ax=ax,
        extent=extent,
        projection=projection,
        background=background,
        zoom=zoom
    )

    # Clip values
    if vmin is not None or vmax is not None:
        data = data.copy()
        data[var] = data[var].clip(vmin, vmax, inplace=True)

    # Add data
    data.plot(column=var, ax=ax, **kwargs)

    # Add annotations
    if annotations is not None:
        if annotation_kwargs is None:
            annotation_kwargs = {}
        add_annotated_points(ax, annotations, **annotation_kwargs)

    # Show the figure
    if show is True:
        plt.show()  # pragma: no cover
    else:
        return fig, ax


def plot_segments(gps_data, *args, **kwargs):
    """Plot segments with background and annotations.

    Args:
        gps_data (:py:class:`~gps_data_analyzer.gps_data._GpsBase`): The object to plot.

    Note:
        The given object is converted into segments then plotted using
        :func:`~gps_data_analyzer.plot_utils.plot`.
    """
    segments = gps_data.segments()

    return plot(segments, *args, **kwargs)



import numpy as np
import pandas as pd

import gps_data_analyzer as gda


def test_simple_rest_time():
    x = [i for i in range(20)]
    y = [0] * 20
    z = [0] * 20
    t = ["2019/06/16-09:55:{:0}".format(i) for i in range(20)]
    df = pd.DataFrame({"x": x, "y": y, "z": z, "t": t})
    data = gda.GpsPoints(df, x_col="x", y_col="y", z_col="z", time_col="t")

    res = gda.time_analysis.compute_rest_time(data, 5)

    assert np.equal(res[:5], range(5, 10)).all()
    assert (res[5:-5] == 10).all()
    assert np.equal(res[-5:], range(9, 4, -1)).all()


def test_rest_time_with_loops():
    x = [i for i in range(11)] + [i for i in range(9, -1, -1)]
    y = [0] * 21
    z = [0] * 21
    t = ["2019/06/16-09:55:{:0}".format(i) for i in range(21)]
    df = pd.DataFrame({"x": x, "y": y, "z": z, "t": t})
    data = gda.GpsPoints(df, x_col="x", y_col="y", z_col="z", time_col="t")

    res = gda.time_analysis.compute_rest_time(data, 5)

    assert np.equal(res[:5], range(5, 10)).all()
    assert res[5] == 20
    assert np.equal(res[6:11], range(18, 9, -2)).all()
    assert np.equal(res[10:15], range(10, 19, 2)).all()
    assert res[15] == 20
    assert np.equal(res[-5:], range(9, 4, -1)).all()



import numpy as np
import pandas as pd
import pyproj
import pytest

import gps_data_analyzer as gda


def test_create_track(simple_gps_data, simple_gps_raw_data):
    x, y, z, t = simple_gps_raw_data

    assert simple_gps_data.x.tolist() == x
    assert simple_gps_data.y.tolist() == y
    assert simple_gps_data.z.tolist() == z
    assert simple_gps_data.t.tolist() == [pd.to_datetime(i) for i in t]


def test_equal_track(simple_gps_data):
    new = gda.GpsPoints(simple_gps_data.data)

    assert new == simple_gps_data


def test_create_track_sort(simple_gps_df, simple_gps_raw_data):
    x, y, z, t = simple_gps_raw_data

    df = simple_gps_df.loc[[2, 0, 1]]
    res = gda.GpsPoints(df, x_col="x", y_col="y", z_col="z", time_col="t")

    assert res.x.tolist() == x
    assert res.y.tolist() == y
    assert np.equal(res.xy, np.vstack((x, y)).T).all()
    assert res.z.tolist() == z
    assert res.t.tolist() == [pd.to_datetime(i) for i in t]
    assert res.crs.to_epsg() == 4326
    assert np.allclose(res.dt.values, [np.nan, 23.0, 135.0], equal_nan=True)
    assert np.allclose(res.dist.values, [np.nan, 15724.02, 15723.75], equal_nan=True)
    assert np.allclose(
        res.velocity.values, [np.nan, 683.6529, 116.4722], equal_nan=True
    )


def test_create_track_proj(simple_gps_df, simple_gps_raw_data):
    x, y, z, t = simple_gps_raw_data

    df = simple_gps_df.loc[[2, 0, 1]]
    res = gda.GpsPoints(
        df, x_col="x", y_col="y", z_col="z", time_col="t", local_crs=2154
    )

    # Compute projected results
    proj = pyproj.Proj(2154)
    xy_proj = [proj(i, j) for i, j in zip(x, y)]
    x_proj = [i[0] for i in xy_proj]
    y_proj = [i[1] for i in xy_proj]

    # Check results
    assert res.x.tolist() == x_proj
    assert res.y.tolist() == y_proj
    assert res.z.tolist() == z
    assert res.t.tolist() == [pd.to_datetime(i) for i in t]
    assert res.crs.to_epsg() == 2154
    assert np.allclose(res.dt.values, [np.nan, 23.0, 135.0], equal_nan=True)
    assert np.allclose(res.dist.values, [np.nan, 20707.888, 20682.199], equal_nan=True)
    assert np.allclose(
        res.velocity.values, [np.nan, 900.343, 153.201], equal_nan=True
    )


def test_add_attribute(simple_gps_data):
    new_attr = simple_gps_data.x * 2
    new_attr.rename("two_x", inplace=True)
    simple_gps_data.add_attribute(new_attr)
    assert (simple_gps_data.two_x.tolist() == (simple_gps_data.x * 2)).all()


def test_add_attribute_name(simple_gps_data):
    new_attr = simple_gps_data.x * 2
    simple_gps_data.add_attribute(new_attr, name="two_x")
    assert (simple_gps_data.two_x.tolist() == (simple_gps_data.x * 2)).all()


def test_poi(simple_poi_data, simple_poi_raw_data):
    x, y, r = simple_poi_raw_data
    assert np.equal(simple_poi_data.x, x).all()
    assert np.equal(simple_poi_data.y, y).all()
    assert np.equal(simple_poi_data.radius, r).all()
    assert simple_poi_data.crs.to_epsg() == 4326


def test_poi_fail(simple_poi_df):
    with pytest.raises(KeyError):
        gda.PoiPoints(simple_poi_df.drop(columns=["radius"]), x_col="x", y_col="y")


def test_mask(simple_poi_data, simple_gps_data):
    assert len(simple_gps_data) == 3

    # Drop from masks
    N = simple_gps_data.drop_from_mask(simple_poi_data)

    # Check results
    assert N == 2
    assert len(simple_gps_data) == 1
    assert np.equal(simple_gps_data.xy, [[0, 1]]).all()
    assert simple_gps_data.index == pd.core.indexes.range.RangeIndex(0, 1, 1)


def test_mask_polygon(simple_poi_data, simple_gps_data):
    assert len(simple_gps_data) == 3

    # Create small polygons aroung PoI points
    polygons = simple_poi_data.buffer(0.01).to_frame("geometry")
    polygons["radius"] = simple_poi_data.radius
    polygons.crs = simple_poi_data.crs

    # Drop from masks
    N = simple_gps_data.drop_from_mask(polygons)

    # Check results
    assert N == 2
    assert len(simple_gps_data) == 1
    assert np.equal(simple_gps_data.xy, [[0, 1]]).all()
    assert simple_gps_data.index == pd.core.indexes.range.RangeIndex(0, 1, 1)


def test_mask_polygon_no_radius(simple_poi_data, simple_gps_data):
    assert len(simple_gps_data) == 3

    # Create small polygons aroung PoI points
    polygons = simple_poi_data.buffer(0.1).to_frame("geometry")
    polygons.crs = simple_poi_data.crs

    # Drop from masks
    N = simple_gps_data.drop_from_mask(polygons)

    # Check results
    assert N == 2
    assert len(simple_gps_data) == 1
    assert np.equal(simple_gps_data.xy, [[0, 1]]).all()
    assert simple_gps_data.index == pd.core.indexes.range.RangeIndex(0, 1, 1)



import numpy as np
import pandas as pd

import gps_data_analyzer as gda


def test_save_load_gps_points(simple_gps_raw_data, simple_gps_data, gpkg_filepath):
    # Save to temporary file
    gda.io.save(simple_gps_data, gpkg_filepath)

    # Load to GpsPoints
    res = gda.load_gps_points(gpkg_filepath)

    # Check results
    x, y, z, t = simple_gps_raw_data
    assert res.x.tolist() == x
    assert res.y.tolist() == y
    assert res.z.tolist() == z
    assert res.datetime.tolist() == [pd.to_datetime(i) for i in t]
    assert np.allclose(res.dt.values, [np.nan, 23.0, 135.0], equal_nan=True)
    assert np.allclose(res.dist.values, [np.nan, 15724.02, 15723.75], equal_nan=True)
    assert np.allclose(
        res.velocity.values, [np.nan, 683.6529, 116.4722], equal_nan=True
    )


def test_save_load_poi_points(simple_poi_raw_data, simple_poi_data, gpkg_filepath):
    # Save to temporary file
    gda.io.save(simple_poi_data, gpkg_filepath)

    # Load to PoiPoints
    poi = gda.load_poi_points(gpkg_filepath)

    # Check results
    x, y, r = simple_poi_raw_data
    assert poi.x.tolist() == x
    assert poi.y.tolist() == y
    assert poi.radius.tolist() == r
    assert "z" not in poi.columns
    assert "datetime" not in poi.columns
    assert "dt" not in poi.columns
    assert "dist" not in poi.columns
    assert "velocity" not in poi.columns



import pytest

import pandas as pd

import gps_data_analyzer as gda


@pytest.fixture
def simple_poi_raw_data():
    x = [0.15]
    y = [1.15]
    r = [0.1]
    return x, y, r


@pytest.fixture
def simple_poi_df(simple_poi_raw_data):
    x, y, r = simple_poi_raw_data
    df = pd.DataFrame({"x": x, "y": y, "radius": r})
    return df


@pytest.fixture
def simple_poi_data(simple_poi_df):
    return gda.PoiPoints(simple_poi_df, x_col="x", y_col="y")



import pandas as pd
import geopandas as gpd


def save(obj, path, mode="w", **kwargs):
    """Save data in a GeoPackage at the given path.

    Args:
        obj (:py:class:`~gps_data_analyzer.gps_data._GpsBase` or ``pandas.DataFrame``):
            The object to save.
        path (str): The path to the GeoPackage.
        kwargs: The given kwargs will be passed to :func:`pandas.DataFrame.to_file`.

    Note:
        If the GeoPackage exists, it is possible to add a new layer with the 'layer'
        argument.
    """
    tmp = obj.data.copy()

    # Convert datetime to string
    if obj._has_time:
        tmp["datetime"] = tmp["datetime"].apply(pd.Timestamp.isoformat)

    # Save to GeoPackage
    tmp.to_file(path, driver="GPKG", encoding="utf-8", **kwargs)

    # TODO: Fiona>=0.19 will be able to store metadata in GPKG files. It would be nice
    # to store the data type in metadata so the load() function can now which class it
    # should call.


def _load(path):
    """Load data from a GeoPackage at the given path.

    Args:
        path (str): The path to the GeoPackage.

    Note:
        This function returns a ``pandas.DataFrame``. Use
        :func:`~gps_data_analyzer.gps_data.load_gps_points` or
        :func:`~gps_data_analyzer.gps_data.load_gps_points` to get a
        :py:class:`~gps_data_analyzer.gps_data.GpsPoints` or a
        :py:class:`~gps_data_analyzer.gps_data.PoiPoints` object.
    """
    data = gpd.read_file(path, driver="GPKG")

    # Convert time columns
    if "datetime" in data.columns:
        data["datetime"] = pd.to_datetime(data["datetime"])

    # If everything could be imported properly, the new object is returned
    return data



from setuptools import find_packages, setup


classifiers = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "Intended Audience :: Education",
    "Intended Audience :: Information Technology",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Topic :: Scientific/Engineering :: GIS",
    "Topic :: Scientific/Engineering",
    "Topic :: Utilities",
]


with open("requirements-install.pip") as f:
    req_base = f.read().splitlines()

with open("requirements-test.pip") as f:
    req_tests = f.read().splitlines()

with open("requirements-dev.pip") as f:
    req_dev = f.read().splitlines()

with open("README.md") as f:
    long_description = f.read()

setup(
    name="GpsDataAnalyzer",
    use_scm_version=True,
    description=(
        "A simple Python toolkit to analyze GPS data"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Adrien Berchet",
    author_email="adrien.berchet@gmail.com",
    url="https://github.com/adrien-berchet/GpsDataAnalyzer",
    license="MIT",
    classifiers=classifiers,
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=False,
    install_requires=req_base,
    extras_require={
        "dev": req_dev + req_tests,
        "tests": req_tests
    },
    setup_requires=["setuptools_scm"],
    python_requires=">=3.6",
    zip_safe=False,
)



import matplotlib.pyplot as plt
import numpy as np
from scipy import spatial

from .plot_utils import add_annotated_points
from .plot_utils import create_transparent_cmap
from .plot_utils import setup_axis


class Extent(object):
    """Class to manage extent of a Raster data and generate a mesh.

    Attributes:
        inner_xmin (float): The minimum X coordinate not considering the border.
        inner_xmax (float): The maximum X coordinate not considering the border.
        inner_ymin (float): The minimum Y coordinate not considering the border.
        inner_ymax (float): The maximum Y coordinate not considering the border.
        xmin (float): The minimum X coordinate considering the border.
        xmax (float): The maximum X coordinate considering the border.
        ymin (float): The minimum Y coordinate considering the border.
        ymax (float): The maximum Y coordinate considering the border.
        border (float): The extra border to add around the data values.

    Args:
        xmin (float): The minimum X coordinate.
        xmax (float): The maximum X coordinate.
        ymin (float): The minimum Y coordinate.
        ymax (float): The maximum Y coordinate.
        border (float, optional): The extra border to add around the data values.
    """

    def __init__(self, xmin, xmax, ymin, ymax, border=0):
        self.border = border
        self.inner_xmin = xmin
        self.inner_xmax = xmax
        self.inner_ymin = ymin
        self.inner_ymax = ymax
        self.xmin = xmin - border
        self.xmax = xmax + border
        self.ymin = ymin - border
        self.ymax = ymax + border

    def reset_border(self, border):
        """Define new extent and recalculate the extent according to it.

        Args:
            border (float): The extra border to add around the data values.
        """
        self.border = border
        self.xmin = self.inner_xmin - border
        self.xmax = self.inner_xmax + border
        self.ymin = self.inner_ymin - border
        self.ymax = self.inner_ymax + border

    def __iter__(self):
        for i in [self.xmin, self.xmax, self.ymin, self.ymax]:
            yield i

    def __getitem__(self, key):
        return [self.xmin, self.xmax, self.ymin, self.ymax][key]

    def mesh(self, mesh_size=None, x_size=None, y_size=None, nx=None, ny=None):
        """Create a mesh in the current extent.

        Args:
            mesh_size (float, optional): The space between two pixels of the heatmap.
            x_size (float, optional): The space between two pixels of the heatmap along
                the X axis.
            y_size (float, optional): The space between two pixels of the heatmap along
                the Y axis.
            nx (int, optional): The number of pixels of the heatmap along the X axis.
            ny (int, optional): The number of pixels of the heatmap along the Y axis.

        Returns:
            (``numpy.array``, ``numpy.array``): X and Y coordinates of the mesh nodes.
        """

        # Check arguments
        if nx is None and ny is None:
            err_msg = (
                "Either 'mesh_size' or both 'x_size' and 'y_size' must be not None"
            )
            if mesh_size is None:
                if x_size is None or y_size is None:
                    raise ValueError(err_msg)
            else:
                if x_size is not None or y_size is not None:
                    raise ValueError(err_msg)

                x_size = y_size = mesh_size

            if x_size <= 0 or y_size <= 0:
                raise ValueError("The mesh size must be > 0")

            nx = complex(0, int(np.round((self.xmax - self.xmin) / x_size)))
            ny = complex(0, int(np.round((self.ymax - self.ymin) / y_size)))
        else:
            if nx is None or ny is None:
                raise ValueError("Both 'nx' and 'ny' must be not None")
            if mesh_size is not None or x_size is not None or y_size is not None:
                raise ValueError(
                    "Either both 'nx' and 'ny' OR 'mesh_size' OR both 'x_size' and "
                    "'y_size' must be not None"
                )
            if nx <= 0 or ny <= 0:
                raise ValueError("Both 'nx' and 'ny' must be > 0")

            nx = complex(0, nx)
            ny = complex(0, ny)

        # Generate mesh (use complex numbers to include the last value)
        X, Y = np.mgrid[self.xmin : self.xmax : nx, self.ymin : self.ymax : ny]

        return X, Y


class Raster(object):
    """Class to manage Raster data.

    Attributes:
        X (int): The default EPSG code of input data (only used when
            ``input_crs`` is ``None``).
        Y (str): The default column name of input data that contains X
            coordinates (only used when ``x_col`` is ``None``).
        values (str): The default column name of input data that contains Y
            coordinates (only used when ``y_col`` is ``None``).
        extent (str): The default column name of input data that contains Z
            coordinates (only used when ``z_col`` is ``None`` and ``_has_z`` is
            ``True``).

    Args:
        X (``numpy.array``): The X coordinates.
        Y (``numpy.array``): The Y coordinates.
        values (``numpy.array``): The values at each point (X, Y).
        extent (:py:class:`~gps_data_analyzer.raster_analysis.Extent`): The extent of
            the raster data.
    """

    def __init__(self, X, Y, values, extent):
        assert X.size == Y.size == values.size
        self.X = X
        self.Y = Y
        self.values = values
        self.extent = extent

    def plot(
        self,
        ax=None,
        show=True,
        projection=None,
        cmap=None,
        background=False,
        zoom=None,
        annotations=None,
        annotation_kwargs=None,
        **kwargs
    ):
        """Plot raster with background and annotations.

        Args:
            ax (``matplotlib.pyplot.Axes``, optional): The axis object to update.
            show (bool, optional): If true, call :func:`plt.show` else return the figure
                and axis objects.
            projection (:py:class:`cartopy.crs.Projection`, optional): The projection of
                the axis (:py:class:`~cartopy.crs.PlateCarree` by default).
            cmap (:py:class:`matplotlib.colors.Colormap`, optional): The colormap to use
                (a default will be created if not given).
            background (bool or :py:class:`cartopy.io.img_tiles.GoogleWTS`, optional): \
                If true, a default background is added using Google Satellite. If a
                :obj:`~cartopy.io.img_tiles.GoogleWTS` object is given, it is used.
            zoom (int, mandatory if :py:obj:`background` is not :py:obj:`None`): The
                zoom value used to generate the background.
            annotations (:py:class:`~gps_data_analyzer.gps_data.PoiPoints`, optional): \
                The points used to annotate the figure.
            annotation_kwargs (dict, optional): The kwargs passed to
                :func:`~gps_data_analyzer.plot_utils.add_annotated_points`.
            kwargs: The given kwargs will be passed to
                :func:`matplotlib.pyplot.Axes.imshow`.

        Returns:
            :py:class:`~gps_data_analyzer.raster_analysis.Raster` The 2D array
            containing the result.
        """

        # Setup axis
        fig, ax = setup_axis(
            ax=ax,
            extent=self.extent,
            projection=projection,
            background=background,
            zoom=zoom
        )

        # Define CMAP
        if cmap is None:
            cmap = create_transparent_cmap()

        # Add raster
        ax.imshow(
            self.values,
            cmap=cmap,
            extent=self.extent,
            origin="upper",
            transform=projection,
            zorder=10,
            **kwargs
        )

        # Add annotations
        if annotations is not None:
            if annotation_kwargs is None:
                annotation_kwargs = {}
            add_annotated_points(ax, annotations, **annotation_kwargs)

        if show is True:
            plt.show()  # pragma: no cover
        else:
            return fig, ax


def heatmap(
    gps_data,
    mesh_size=None,
    x_size=None,
    y_size=None,
    nx=None,
    ny=None,
    border=0,
    kernel_size=None,
    kernel_cut=4.0,
    weight_col=None,
    normalize=True,
):
    """Compute heatmap from a set of points using a Gaussian kernel.

    Args:
        gps_data (:py:class:`~gps_data_analyzer.gps_data._GpsBase`): The point set.
        mesh_size (float, optional): The space between two pixels of the heatmap.
        x_size (float, optional): The space between two pixels of the heatmap along the
            X axis.
        y_size (float, optional): The space between two pixels of the heatmap along the
            Y axis.
        nx (int, optional): The number of pixels of the heatmap along the X axis.
        ny (int, optional): The number of pixels of the heatmap along the Y axis.
        border (float, optional): The extra border around the data.
        kernel_size (float, optional): The kernel size used for computation.
        kernel_cut (float, optional): The kernel cut used for computation.
        weight_col (str, optional): The column name used as point weights.
        normalize (bool, optional): Trigger normalization of the result.

    Returns:
        :py:class:`~gps_data_analyzer.raster_analysis.Raster` The 2D array containing
        the result.
    """
    # Check arguments
    if kernel_size is not None and kernel_size <= 0:
        raise ValueError("The 'kernel_size' argument must be > 0")
    if kernel_cut is not None and kernel_cut <= 0:
        raise ValueError("The 'kernel_cut' argument must be > 0")

    # Get coordinates
    x = gps_data.x
    y = gps_data.y
    if weight_col is not None:
        weight = getattr(gps_data, weight_col).values
    else:
        weight = np.ones(len(gps_data))

    # Compute extent
    xmin = x.min()
    xmax = x.max()
    ymin = y.min()
    ymax = y.max()
    extent = Extent(xmin, xmax, ymin, ymax, border)

    # Generate mesh
    X, Y = extent.mesh(mesh_size=mesh_size, x_size=x_size, y_size=y_size, nx=nx, ny=ny)
    positions = np.vstack([X.ravel(), Y.ravel()])

    # Compute KDTree
    tree = spatial.KDTree(positions.T)

    # Init sigma if not given
    if kernel_size is None:
        if X.size >= 4:
            dx = X[1, 0] - X[0, 0]
        else:
            dx = 0
        if Y.size >= 4:
            dy = Y[0, 1] - Y[0, 0]
        else:
            dy = 0
        mesh_size = max(dx, dy)
        if mesh_size > 0:
            kernel_size = 2.0 * mesh_size
        else:
            kernel_size = 1

    kde = np.zeros(len(tree.data))
    for num, (_x, _y, _w) in enumerate(zip(x, y, weight)):  # TODO: optimize this loop
        # Get the closest points of the current point
        coords_i = [_x, _y]
        in_radius_pts = tree.query_ball_point(coords_i, kernel_cut * kernel_size)

        # Compute distances and divide by the krenel size
        q = (
            np.squeeze(spatial.distance.cdist(tree.data[in_radius_pts], [coords_i]))
            / kernel_size
        )

        # Compute KDE contribution
        res = np.exp(-np.power(q, 2)) / (2.0 * kernel_size)
        kde[in_radius_pts] += res * _w

    # Normalize KDE
    if not normalize:
        kde /= kernel_size * np.sqrt(2 * np.pi)
    else:
        kde -= kde.min()
        kde /= kde.max()

    # Reshape and rotate the result
    heatmap = np.rot90(np.reshape(kde, X.shape))

    return Raster(X, Y, heatmap, extent)



import pytest
import types

import cartopy.io.img_tiles as cimgt
import numpy as np

import gps_data_analyzer as gda


def GOOGLE_IMAGE_URL_REPLACEMENT(self, tile):
    url = ('https://chart.googleapis.com/chart?chst=d_text_outline&'
           'chs=256x256&chf=bg,s,00000055&chld=FFFFFF|16|h|000000|b||||'
           'Google:%20%20(' + str(tile[0]) + ',' + str(tile[1]) + ')'
           '|Zoom%20' + str(tile[2]) + '||||||______________________'
           '______')
    return url


@pytest.fixture
def background():
    """
    This comes from the tests of the Cartopy library. It ensures that the background
    will never be changed by google and is only used in tests.
    """
    gt = cimgt.GoogleTiles()
    gt._image_url = types.MethodType(GOOGLE_IMAGE_URL_REPLACEMENT, gt)
    return gt


@pytest.mark.mpl_image_compare
def test_plot(simple_gps_data):
    # Simple plot
    fig, ax = gda.plot_utils.plot(simple_gps_data, var="z", show=False)
    return fig


@pytest.mark.mpl_image_compare
def test_plot_border(simple_gps_data):
    # Plot with border
    fig, ax = gda.plot_utils.plot(simple_gps_data, var="z", border=0.02, show=False)
    return fig


@pytest.mark.mpl_image_compare
def test_plot_cmap(simple_gps_data):
    # Plot with cmap
    fig, ax = gda.plot_utils.plot(
        simple_gps_data, var="z", border=0.02, cmap="Spectral", show=False)
    return fig


@pytest.mark.mpl_image_compare
def test_plot_annotations(simple_gps_data, simple_poi_data):
    # Plot with cmap
    fig, ax = gda.plot_utils.plot(
        simple_gps_data,
        var="z",
        border=0.02,
        show=False,
        annotations=simple_poi_data)
    return fig


@pytest.mark.mpl_image_compare
def test_plot_background(simple_gps_data, background):
    # Plot with background
    fig, ax = gda.plot_utils.plot(
        simple_gps_data,
        var="z",
        background=background,
        zoom=11,
        border=0.02,
        show=False)
    return fig


@pytest.mark.mpl_image_compare
def test_plot_segments(simple_gps_data):
    # Plot segments
    fig, ax = gda.plot_utils.plot_segments(
        simple_gps_data,
        var="velocity",
        show=False
    )
    return fig


@pytest.mark.mpl_image_compare
def test_plot_raster(simple_gps_data):
    # Define extent
    extent = gda.raster_analysis.Extent(0, 1, 0, 1, 0.1)

    # Define raster
    X, Y = extent.mesh(nx=5, ny=5)
    values = np.reshape(np.arange(0, 25), X.shape)
    raster = gda.raster_analysis.Raster(X, Y, values, extent)

    # Plot raster
    fig, _ = raster.plot(show=False)

    return fig


@pytest.mark.mpl_image_compare
def test_plot_heatmap(simple_gps_data):
    # Define raster
    raster = gda.raster_analysis.heatmap(
        simple_gps_data, weight_col="z", nx=15, ny=15, border=0.05)

    # Plot raster
    fig, ax = raster.plot(show=False)

    # Add track markers
    x = simple_gps_data.x
    y = simple_gps_data.y
    ax.plot(x, y, "k.", markersize=10, transform=ax.projection, zorder=15)

    return fig


@pytest.mark.mpl_image_compare
def test_plot_heatmap_annotations(simple_gps_data, simple_poi_data):
    # Define raster
    raster = gda.raster_analysis.heatmap(
        simple_gps_data, weight_col="z", nx=15, ny=15, border=0.05)

    # Plot raster and annotations
    fig, ax = raster.plot(show=False, annotations=simple_poi_data)

    # Add track markers
    x = simple_gps_data.x
    y = simple_gps_data.y
    ax.plot(x, y, "k.", markersize=10, transform=ax.projection, zorder=15)

    return fig



import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import LineString

from .utils import haversine
from . import io


DEFAULT_TIME_FORMAT = "%Y/%m/%d-%H:%M:%S"


def _convert_time(series, format=DEFAULT_TIME_FORMAT):
    """Convert a ``pandas.Series`` containing timestamps as strings.

    Args:
        series (pandas.Series): The timestamps given as strings.
        format (str, optional): The format used for conversion.

    Returns:
        ``pandas.Series``: The timestamps converted in ``pandas.datetime``.
    """
    return pd.to_datetime(series, format=format)


class _GpsBase(object):
    """Class to wrap a ``geopandas.GeoDataFrame`` and format it in order to store GPS
    points.

    Attributes:
        _default_input_crs (int): The default EPSG code of input data (only used when
            ``input_crs`` is ``None``).
        _default_x_col (str): The default column name of input data that contains X
            coordinates (only used when ``x_col`` is ``None``).
        _default_y_col (str): The default column name of input data that contains Y
            coordinates (only used when ``y_col`` is ``None``).
        _default_z_col (str): The default column name of input data that contains Z
            coordinates (only used when ``z_col`` is ``None`` and ``_has_z`` is
            ``True``).
        _default_time_col (str): The default column name of input data that contains
            timestamps (only used when ``time_col`` is ``None`` and ``_has_time`` is
            ``True``).
        _has_z (bool): Indicate whether the Z coordinate must be considered or not.
        _has_time (bool): Indicate whether the timestamps must be considered or not.
        _use_haversine (bool): Indicate whether the distance computations must use the
            Haversine formula or not.
        datetime_format (str): The format used to convert strings into timestamps.

    Args:
        df (``pandas.DataFrame``): The data that will be formatted and stored.
        input_crs (int): The EPSG code of the input data.
        local_crs (int): The EPSG code of the local projection to which the data will
            be transformed.
        keep_cols (:obj:`list` of :obj:`str`): The columns that should not be discarded.
        x_col (str): The name of the column that contains X coordinates.
        y_col (str): The name of the column that contains Y coordinates.
        z_col (str): The name of the column that contains Z coordinates.
        time_col (str): The name of the column that contains timestamps.
    """

    _default_input_crs = 4326
    _default_x_col = "x"
    _default_y_col = "y"
    _default_z_col = "z"
    _default_time_col = "datetime"
    _has_z = True
    _has_time = False
    _use_haversine = True

    datetime_format = DEFAULT_TIME_FORMAT

    def __init__(
        self,
        df,
        input_crs=None,
        local_crs=None,
        keep_cols=None,
        x_col=None,
        y_col=None,
        z_col=None,
        time_col=None,
    ):

        input_crs = input_crs if input_crs is not None else self._default_input_crs
        local_crs = local_crs if local_crs is not None else input_crs
        x_col = x_col if x_col is not None else self._default_x_col
        y_col = y_col if y_col is not None else self._default_y_col
        z_col = z_col if z_col is not None else self._default_z_col
        time_col = time_col if time_col is not None else self._default_time_col

        # Format data
        gdf = self._format_data(
            df,
            input_crs,
            local_crs,
            x_col,
            y_col,
            z_col,
            time_col,
            keep_cols=keep_cols,
        )

        # Save data
        self.data = gdf
        self.crs = self.data.crs

        if self._has_time:
            self._normalize_data()

    @property
    def x(self):
        """``pandas.Series``: Get X coordinates from the geometry."""
        return self.data.geometry.x

    @property
    def y(self):
        """``pandas.Series``: Get Y coordinates from the geometry."""
        return self.data.geometry.y

    @property
    def t(self):
        """Get timestamps if :py:attr:`~_has_time` is ``True`` or the `t` column
        otherwise.

        Returns:
            ``pandas.Series``: Timestamps.
        """
        if self._has_time:
            attr = "datetime"
        else:
            attr = "t"
        return self.__getattr__(attr)

    @property
    def xy(self):
        """``numpy.array``: Array with a (x,y) couple of each point."""
        return np.vstack([self.data.geometry.x, self.data.geometry.y]).T

    def __eq__(self, other):
        """Compare two :py:obj:`~_GpsBase` objects"""
        assert isinstance(other, _GpsBase), (
            "The operator == is only defined for " "'_GpsBase' objects."
        )
        return self.data.equals(other.data)

    def __iter__(self):
        """Return a generator over the rows of the internal
        ``geopandas.GeoDataFrame``"""
        for i in self.data.iterrows():
            yield i

    def __getattr__(self, attr):
        """Apply any unknown attribute to the internal ``geopandas.GeoDataFrame``."""
        return getattr(self.data, attr)

    def __len__(self):
        """Get the length of the internal data.

        Returns:
            int: The length of the internal ``geopandas.GeoDataFrame``
        """
        return len(self.data)

    def _format_data(
        self,
        df,
        input_crs,
        local_crs,
        x_col,
        y_col,
        z_col=None,
        time_col=None,
        keep_cols=None,
    ):
        """Format a ``pandas.DataFrame`` or ``geopandas.GeoDataFrame``.

        Args:
            df (``pandas.DataFrame`` or ``geopandas.GeoDataFrame``): The object to
                format.
            input_crs (int): The EPSG code of the input data.
            local_crs (int or None): The EPSG code of the local projection to which the
                data will be transformed.
            x_col (str): The name of the column containing X or lon coordinates.
            y_col (str): The name of the column containing Y or lat coordinates.
            z_col (str, optional): The name of the column containing Z coordinates.
            time_col (str, optional): The name of the column containing timestamps.
            keep_cols (:obj:`list` of :obj:`str`, optional): The names of the columns
                that should be kept (all others will be discarded).

        Note:
            The column containing timestamps can be either in string format (and should
            thus follow the format given by :py:attr:`~datetime_format`) or in a subtype
            of ``numpy.datetime64``.

        Returns:
            ``geopandas.GeoDataFrame``: The formatted data.
        """
        df = df.copy()

        # Convert time and sort by time
        if self._has_time:
            t_col = df[time_col]
            if not np.issubdtype(df[time_col].dtype, np.datetime64):
                t_col = _convert_time(df[time_col], format=self.datetime_format)
            df["datetime"] = t_col
            df.sort_values("datetime", inplace=True)

        if not isinstance(df, gpd.GeoDataFrame):
            # Drop missing coordinates
            df.dropna(subset=[x_col, y_col], inplace=True)

            # Convert to GeoDataFrame
            df = gpd.GeoDataFrame(
                df, crs=input_crs, geometry=gpd.points_from_xy(df[x_col], df[y_col])
            )
        else:
            input_crs = df.crs.to_epsg()

        # Drop useless columns
        cols = ["geometry"]
        if self._has_z:
            cols.append(z_col)
        if self._has_time:
            cols.append("datetime")
        df = df[cols + ([] if keep_cols is None else keep_cols)]

        # Normalize column names
        if self._has_z:
            df.rename(columns={z_col: "z"}, inplace=True)

        # Reset index
        df.reset_index(drop=True, inplace=True)

        # Project data
        if local_crs is not None and local_crs != input_crs:
            df.to_crs(local_crs, inplace=True)

        return df

    def _normalize_data(self):
        """Conpute time delta between consecutive points (in s)."""
        self.data["dt"] = (
            self.data["datetime"] - self.data["datetime"].shift()
        ).values / pd.Timedelta(1, "s")

        # Conpute distance between consecutive points (in m)
        shifted = self.data.geometry.shift()
        if self.crs.to_epsg() == 4326 and self._use_haversine:
            self.data["dist"] = haversine(self.y, self.x, shifted.y, shifted.x)
        else:
            self.data["dist"] = self.data.distance(shifted)

        # Conpute velocity between consecutive points (in m/s)
        self.data["velocity"] = self.data["dist"] / self.data["dt"]

    def add_attribute(self, attr, name=None):
        """Add a column to the internal ``geopandas.GeoDataFrame``.

        Args:
            attr (``pandas.Series``): The column to add.
            name (str, optional): The name of the new attribute. If not provided, the
                name of the ``pandas.Series`` is used.

        Note:
            The labels of the given ``pandas.Series`` must be the same as the ones of
            the internal ``geopandas.GeoDataFrame``.
        """
        assert isinstance(attr, pd.Series), (
            "The 'attr' argument must be a" "pandas.Series"
        )
        if name is not None:
            self.data[name] = attr
        else:
            self.data[attr.name] = attr

    def segments(self):
        """Build segments from the consecutive points.

        Returns:
            ``geopandas.GeoDataFrame``: A ``geopandas.GeoDataFrame`` containing the
            segments.
        """
        tmp = (
            self.data[["geometry"]]
            .join(self.data[["geometry"]].shift(), rsuffix="_m1")
            .dropna()
        )
        lines = tmp.apply(
            axis=1, func=lambda x: LineString([x.geometry_m1, x.geometry])
        )
        lines.name = "geometry"
        segments = self.data[["dt", "dist", "velocity"]].join(lines, how="right")
        return gpd.GeoDataFrame(segments, crs=self.crs, geometry="geometry")

    def drop_from_mask(self, mask):
        """Drop points contained in the given mask.

        Args:
            mask (:obj:`geopandas.GeoDataFrame`): The mask used to drop internal points.

        Note:
            * The mask must be a :py:obj:`_GpsBase` or ``geopandas.GeoDataFrame``
              object.
            * If the mask has a `radius` column, it will be used and drop all points at
              a distance smaller than the `radius` values.

        Returns:
            int: The number of dropped points.
        """
        mask = mask.copy()

        if isinstance(mask, pd.Series):
            mask = gpd.GeoDataFrame(mask.to_frame("geometry"), crs=mask.crs)

        # Project the mask if needed
        if self.crs is not None:
            mask = mask.to_crs(self.crs)

        # Get the points included in masks
        in_mask_pts = pd.Series(np.zeros(len(self)), dtype=bool)
        for num, i in mask.iterrows():
            in_mask_pts = in_mask_pts | (
                self.geometry.distance(i.geometry) <= i.get("radius", 0))

        # Count the number of points that are going to be dropped
        N = in_mask_pts.sum()

        # Drop points in mask
        self.data.drop(in_mask_pts.loc[in_mask_pts].index, inplace=True)
        self.data.reset_index(drop=True, inplace=True)

        return N


class GpsPoints(_GpsBase):
    """Class to store GPS points with Z coordinates and timestamps."""
    _has_z = True
    _has_time = True


def load_gps_points(path):
    """Load :py:obj:`GpsPoints` from a file.

    Args:
        path (str): The path to the file.

    Returns:
        :py:obj:`GpsPoints`: The data loaded.
    """
    return GpsPoints(io._load(path))


class PoiPoints(_GpsBase):
    """Class to store PoI points with only X and Y coordinates."""
    _has_z = False
    _has_time = False

    def __init__(self, *args, **kwargs):
        if len(args) >= 3:
            keep_cols = args[3]
        elif "keep_cols" in kwargs:
            keep_cols = kwargs["keep_cols"]
        else:
            keep_cols = []
            kwargs["keep_cols"] = keep_cols

        keep_cols.extend(["radius"])
        super().__init__(*args, **kwargs)


def load_poi_points(path):
    """Load :py:obj:`PoiPoints` from a file.

    Args:
        path (str): The path to the file.

    Returns:
        :py:obj:`PoiPoints`: The data loaded.
    """
    return PoiPoints(io._load(path))
